//
//  GMlib -- Graphics & Media Lab Common Source Library
//
//  $Id: gmlglrc.cpp,v 1.6 2004/04/02 15:16:08 boki Exp $
//
//  Copyright (C) 2004, Moscow State University Graphics & Media Lab
//  gmlsupport@graphics.cs.msu.su
//  
//  This file is part of GMlib software.
//  For conditions of distribution and use, see the accompanying README file.

/* 

  @file
  @brief Implementation of GLRC
  
*/

#include "../base/gmlcommon.h"
#include <windows.h>
#include "gmlglrc.h"


using gml::GLRC;

//////////////////////////////////////////////////////////////////////////////
/// @param wnd window handle for which gl context should be created
/// 
GLRC::GLRC(HWND wnd)
  :  m_created  ( false )
  ,  m_bMultisamleChecked ( false )
{
  ASSERT( wnd );
  m_wnd = wnd;
  m_dc = ::GetDC( wnd );
  ASSERT( m_dc );
}

//////////////////////////////////////////////////////////////////////////////
///
GLRC::GLRC( HDC hdc )
  : m_created  ( false )
  , m_bMultisamleChecked ( false )
  , m_wnd      ( 0 )
{
  m_dc = hdc;
  ASSERT( m_dc );
}

//////////////////////////////////////////////////////////////////////////////
/// @note Destroy() called inside destructor if was not called already
GLRC::~GLRC()
{
  if (m_created)
    Destroy();
}


//////////////////////////////////////////////////////////////////////////////
///
void GLRC::Destroy()
{
  wglDeleteContext(m_glrc);
  ::ReleaseDC(m_wnd, m_dc);
  m_created = false;
}

//////////////////////////////////////////////////////////////////////////////
///
bool GLRC::MakeCurrent()
{
  ASSERT( m_created );

  if (IsCurrent())
    return true;

  VERIFY(wglMakeCurrent(m_dc, m_glrc));

  return true;
}

//////////////////////////////////////////////////////////////////////////////
/// @note After creation, gl context is automatically become current
/// @return false in case of failure, true otherwise
bool GLRC::Create()
{
  ASSERT( !m_created );

  int nPixelFormat = 0;

  DWORD flags;
  flags = 	PFD_DRAW_TO_WINDOW |									// Format Must Support Window
			PFD_SUPPORT_OPENGL |									// Format Must Support OpenGL
			PFD_DOUBLEBUFFER;										// Must Support Double Buffering

  static PIXELFORMATDESCRIPTOR pfd =
  {
	sizeof (PIXELFORMATDESCRIPTOR),									// Size Of This Pixel Format Descriptor
	1,																// Version Number
	flags,
	PFD_TYPE_RGBA,													// Request An RGBA Format
	24,																// Select Our Color Depth
	0, 0, 0, 0, 0, 0,												// Color Bits Ignored
	1,																// Alpha Buffer
	0,																// Shift Bit Ignored
	0,																// No Accumulation Buffer
	0, 0, 0, 0,														// Accumulation Bits Ignored
	32,																// 16Bit Z-Buffer (Depth Buffer)  
	0,																// No Stencil Buffer
	0,																// No Auxiliary Buffer
	PFD_MAIN_PLANE,													// Main Drawing Layer
	0,																// Reserved
	0, 0, 0															// Layer Masks Ignored
  };

  pfd.cAlphaBits = 8;

 
  nPixelFormat = ChoosePixelFormat( m_dc, &pfd );


  VERIFY( SetPixelFormat( m_dc, nPixelFormat, &pfd ) );

  m_glrc = wglCreateContext( m_dc );

  m_created = true;
 
  return MakeCurrent();
}



//////////////////////////////////////////////////////////////////////////////
/// @note After creation, gl context is automatically become current
/// @return false in case of failure, true otherwise
bool GLRC::CreateMultisampled( int iMultisampleMode )
{
  ASSERT( !m_created );
  
  if( iMultisampleMode == 0 )
    return Create();

  DWORD flags;
  flags = 	PFD_DRAW_TO_WINDOW |									// Format Must Support Window
			PFD_SUPPORT_OPENGL |									// Format Must Support OpenGL
			PFD_DOUBLEBUFFER;										// Must Support Double Buffering

  static PIXELFORMATDESCRIPTOR pfd =
  {
	sizeof (PIXELFORMATDESCRIPTOR),									// Size Of This Pixel Format Descriptor
	1,																// Version Number
	flags,
	PFD_TYPE_RGBA,													// Request An RGBA Format
	24,																// Select Our Color Depth
	0, 0, 0, 0, 0, 0,												// Color Bits Ignored
	1,																// Alpha Buffer
	0,																// Shift Bit Ignored
	0,																// No Accumulation Buffer
	0, 0, 0, 0,														// Accumulation Bits Ignored
	32,																// 16Bit Z-Buffer (Depth Buffer)  
	0,																// No Stencil Buffer
	0,																// No Auxiliary Buffer
	PFD_MAIN_PLANE,													// Main Drawing Layer
	0,																// Reserved
	0, 0, 0															// Layer Masks Ignored
  };

  pfd.cAlphaBits = 8;


  VERIFY( SetPixelFormat( m_dc, iMultisampleMode, &pfd ) );

  m_glrc = wglCreateContext( m_dc );

  m_created = true;

  return MakeCurrent();
}

//////////////////////////////////////////////////////////////////////////////
///
void GLRC::SwapBuffers()
{
  ASSERT(m_created);
  ::SwapBuffers(m_dc);
}

//////////////////////////////////////////////////////////////////////////////
///
bool GLRC::IsCurrent() const
{
  ASSERT( m_created );
  return ::wglGetCurrentContext() == m_glrc;
}


//////////////////////////////////////////////////////////////////////////////
/// @return false in case context is not created, true otherwise
bool GLRC::IsCreated()
{
  return m_created;
}

//////////////////////////////////////////////////////////////////////////////
/// @note Checks for availability of multisampling. 
/// @return false in case multisample is not supported, true otherwise
bool GLRC::IsMultisampleSupported()
{
	return InitMultisample( m_wnd );
}

//////////////////////////////////////////////////////////////////////////////
/// @return structure containing flags indicating availability of multisample modes
gml::sMultisampleModes GLRC::GetMultisampleModes()
{
  ASSERT( m_bMultisamleChecked );

	return m_sAvailableModes;
}

//////////////////////////////////////////////////////////////////////////////
/// WGLisExtensionSupported: This Is A Form Of The Extension For WGL
bool GLRC::WGLisExtensionSupported( const char *pExtension )
{
	const size_t extlen = strlen(pExtension);
	const char *supported = NULL;

	// Try To Use wglGetpExtensionStringARB On Current DC, If Possible
	PROC wglGetExtString = wglGetProcAddress("wglGetExtensionsStringARB");

	if (wglGetExtString)
		supported = ((char*(__stdcall*)(HDC))wglGetExtString)(wglGetCurrentDC());

	// If That Failed, Try Standard Opengl Extensions String
	if (supported == NULL)
		supported = (char*)glGetString(GL_EXTENSIONS);

	// If That Failed Too, Must Be No Extensions Supported
	if (supported == NULL)
		return false;

	// Begin Examination At Start Of String, Increment By 1 On False Match
	for (const char* p = supported; ; p++)
	{
		// Advance p Up To The Next Possible Match
		p = strstr(p, pExtension);

		if (p == NULL)
			return false;															// No Match

		// Make Sure That Match Is At The Start Of The String Or That
		// The Previous Char Is A Space, Or Else We Could Accidentally
		// Match "wglFunkywglExtension" With "wglExtension"

		// Also, Make Sure That The Following Character Is Space Or NULL
		// Or Else "wglExtensionTwo" Might Match "wglExtension"
		if ((p==supported || p[-1]==' ') && (p[extlen]=='\0' || p[extlen]==' '))
			return true;															// Match
	}
}

//////////////////////////////////////////////////////////////////////////////
/// InitMultisample: Used To Query The Multisample Frequencies
bool GLRC::InitMultisample( HWND hWnd )
{  
	// See If The String Exists In WGL!
	if ( !WGLisExtensionSupported( "WGL_ARB_multisample" ) )
	{
		m_sAvailableModes.m_bMultisampleSupported = false;
		return false;
	}

	// Get Our Pixel Format
	PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = 
		(PFNWGLCHOOSEPIXELFORMATARBPROC)
			wglGetProcAddress( "wglChoosePixelFormatARB" );	
	
	if ( !wglChoosePixelFormatARB ) 
	{
		m_sAvailableModes.m_bMultisampleSupported = false;
		return false;
	}

	// Get Our Current Device Context
	HDC hDC = GetDC( hWnd );

	int		pixelFormat;
	int		valid;
	UINT	numFormats;
	float	fAttributes[] = { 0, 0 };

	// These Attributes Are The Bits We Want To Test For In Our Sample
	// Everything Is Pretty Standard, The Only One We Want To 
	// Really Focus On Is The SAMPLE BUFFERS ARB And WGL SAMPLES
	// These Two Are Going To Do The Main Testing For Whether Or Not
	// We Support Multisampling On This Hardware.
	int iAttributes[] =
	{
		WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
		WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
		WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB, 24,
		WGL_ALPHA_BITS_ARB, 8,
		WGL_DEPTH_BITS_ARB, 16,
		WGL_STENCIL_BITS_ARB, 0,
		WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
		WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,
		WGL_SAMPLES_ARB, 2,
		0, 0
	};

	// Test multi = 2
	
	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatARB( hDC , iAttributes , fAttributes , 1 , &pixelFormat , &numFormats );
	
	// If We Returned True, And Our Format Count Is Greater Than 1
	if ( valid && numFormats >= 1 )
	{
    m_sAvailableModes.m_bMultisampleSupported = true;
		m_sAvailableModes.m_bMulti2x = true;
    m_sAvailableModes.m_iMulti2x = pixelFormat;
	}

	// Test multi = 4
	iAttributes[19] = 4;

	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatARB( hDC , iAttributes , fAttributes , 1 , &pixelFormat , &numFormats );
	
	// If We Returned True, And Our Format Count Is Greater Than 1
	if ( valid && numFormats >= 1 )
	{
		m_sAvailableModes.m_bMultisampleSupported = true;
		m_sAvailableModes.m_bMulti4x = true;
    m_sAvailableModes.m_iMulti4x = pixelFormat;
 
	}

	// Test multi = 6
	iAttributes[19] = 6;

	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatARB( hDC , iAttributes , fAttributes , 1 , &pixelFormat , &numFormats );
	
	// If We Returned True, And Our Format Count Is Greater Than 1
	if ( valid && numFormats >= 1 )
	{
		m_sAvailableModes.m_bMultisampleSupported = true;
		m_sAvailableModes.m_bMulti6x = true;	
    m_sAvailableModes.m_iMulti6x = pixelFormat;
	}
	
  m_bMultisamleChecked = true;

	// Return The Valid Format
	return  m_sAvailableModes.m_bMultisampleSupported;
}
