//
//  GMlib -- Graphics & Media Lab Common Source Library
//
//  $Id: gmlglrc.h,v 1.6 2004/04/03 14:19:27 leo Exp $
//
//  Copyright (C) 2004, Moscow State University Graphics & Media Lab
//  gmlsupport@graphics.cs.msu.su
//  
//  This file is part of GMlib software.
//  For conditions of distribution and use, see the accompanying README file.

/** @file 
*  @brief Class for operations with OpenGL Win32 context
* 
*/

#ifndef _GMLGLRC_H_
#define _GMLGLRC_H_

// OpenGL includes
#include <gl/gl.h>
#include <gl/glu.h>

// OpenGL Extensions include
#include "wglext.h"		//WGL extensions
#include "glext.h"		//GL extensions

#include <vector>

// Multisample declarations
#define WGL_SAMPLE_BUFFERS_ARB		 0x2041
#define WGL_SAMPLES_ARB			     0x2042


namespace gml
{
  /** 
  @addtogroup UtilsGL
  @{
  */

  /**
  @struct sMultisampleModes
  @brief  Structure used for multisample handling
  Usage:

  Multisample enabling algorithm is two-pass! 

  On the first pass we create window and context as usual, and check for availability of multisampling.
  On the second pass we create window and then create multisampled context. 
  
  @attention You Must destroy window before second pass, and then recreate it.

  @note Don't forget to enable/disable multisampling: glEnable( GL_MULTISAMPLE_ARB );
 
  @code
  // We must store checked modes not inside wnd
  sMultisampleModes sAvailableModes;

  // First Pass

  //...

  // Create window..

  // Create glrc as usual
  GLRC glrc(wnd);
  glrc.Create();
  
  //...
  
  if ( glrc.IsMultisampleSupported() )
  {
    sAvailableModes = glrc.GetMultisampleModes();
  }

  // End of first pass

  // So we have available modes, and their pixel formats
  // We choose some format, for example Multi2x. 
  // We must check if it is supported, and if it is we can 
  // render with Multisample 2x
  
  // Second pass
  
  // ...
   
  // Create window
  // Create multisampled glrc
  GLRC glrc(wnd)
  glrc.CreateMultisampled( sAvailableModes.SetMultisample2x() );
  
  // ...

  glEnable( GL_MULTISAMPLE_ARB );

  // RENDER HERE

  glDisable( GL_MULTISAMPLE_ARB );

  // ...

  @endcode

  */
  struct sMultisampleModes
  {
    /// Constructor
    sMultisampleModes()
      : m_bMultisampleSupported ( false )
      , m_bMulti2x ( false )
      , m_bMulti4x ( false )
      , m_bMulti6x ( false )
      , m_iMulti2x ( 0 )
      , m_iMulti4x ( 0 )
      , m_iMulti6x ( 0 )
    {
    }

    bool  m_bMultisampleSupported,      ///< True if Multisampling is supported
          m_bMulti2x,                   ///< True if Multisampling 2x is supported
          m_bMulti4x,                   ///< True if Multisampling 4x is supported   
          m_bMulti6x;                   ///< True if Multisampling 6x is supported

    int   m_iMulti2x,                   ///< Pixel Format number for Multisampling 2x
          m_iMulti4x,                   ///< Pixel Format number for Multisampling 4x
          m_iMulti6x;                   ///< Pixel Format number for Multisampling 6x
  
    /// @name Multisample setting members
    /// @{

    /// Set Multisample 2x
    /// @return Multisample 2x pixel format value
    int SetMultisample2x()
    {
      if( m_bMultisampleSupported && m_bMulti2x )
        return m_iMulti2x;
      else 
        return 0;
    }

    /// Set Multisample 4x
    /// @return Multisample 4x pixel format value
    int SetMultisample4x()
    {
      if( m_bMultisampleSupported && m_bMulti4x )
        return m_iMulti4x;
      else 
        return 0;
    }
    
    /// Set Multisample 6x
    /// @return Multisample 6x pixel format value
    int SetMultisample6x()
    {
      if( m_bMultisampleSupported && m_bMulti6x )
        return m_iMulti6x;
      else 
        return 0;
    }
    /// @}
  };


  /**
  @class GLRC
  @brief OpenGL context handling
  @attention Include windows.h before using this file!
  Usage:
  @code
  GLRC glrc(wnd);
  // inside window creation
  glrc.Create();
  // before rendering into window
  glrc.MakeCurrent();
  // ... 
  // display in the screen (in case of double buffering)
  glrc.SwapBuffers();
  // ..
  // delete gl context (may be avoided, because called in destructor anyway)
  glrc.Destroy();
  @endcode
  @todo Add possibility to select pixel format (double/single buffer, stencil, etc)
  */
  class GLRC
  {
  public:
    

    /// @name Class Construction/Destruction
    /// @{
    GLRC( HWND wnd );       ///< Constructor from Window Handle (HWND)
    GLRC( HDC hdc  );      ///< Constructor form Device Context Handle(HDC) 
    
    /// Destory GL context (also called in destructor)
    void Destroy();
    
    virtual ~GLRC();    ///< Virtual Destructor
    /// @}
    
    /// @name Context members
    /// @{
    
    /// Member, needed to create context. Call this before use of OpenGL
    virtual bool Create();
    
    /// Member, used to check is OpenGL Context ready for work.
    bool IsCreated();
    
    /// Member, used for creation of multisampled context, call it after window is created.
    bool CreateMultisampled( int iMultsampleMode );
    
    /// Member, that return true if opengl is ready to draw
    bool IsCurrent() const;
    
    /// Prepare current context for use
    bool MakeCurrent();
    
    /// Call this insted of opengl's swap buffer, at the end of the drawing
    void SwapBuffers();

    /// @}
    
    /// @name Multisample handling
    /// @{

    /// Member, used for checking on availability of multisamling, and multisampling modes
    bool IsMultisampleSupported();
    
    /// Member, used to get available multisample modes
    sMultisampleModes GetMultisampleModes();
    
    /// @}

  private:
    /// @name Multisample checking members
    /// @{
    
    /// Check if an extension is suported
    bool WGLisExtensionSupported( const char *pExtension );
    
    /// Used To Query The Multisample Frequencies
    bool InitMultisample( HWND hWnd );
    
    /// @}
	protected:

	  /// @name Context params
	  /// @{
      bool m_created;					///< true inside Create() / Destory() pair
      HWND m_wnd;						  ///< Window handle
      HDC m_dc;							  ///< Device Context handle
      HGLRC m_glrc;						///< OpenGL Context handle
      /// @}

	  /// @name Multisample params
	  /// @{
    bool m_bMultisamleChecked;            ///< true if multisample check is done
    sMultisampleModes m_sAvailableModes;  ///< Available multisample modes
	  /// @}
  };



  /** @} */
} // namespace gml

#endif